# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import re

from .exceptions import AmcrestError
from .utils import to_unit, percent

_USED = '.UsedBytes'
_TOTAL = '.TotalBytes'


def _express_as(value, unit):
    try:
        return to_unit(value, unit)
    except (TypeError, ValueError):
        return 'unknown', unit


class Storage(object):

    @property
    def storage_device_info(self):
        ret = self.command(
            'storageDevice.cgi?action=getDeviceAllInfo'
        )
        return ret.content.decode('utf-8')

    @property
    def storage_device_names(self):
        ret = self.command(
            'storageDevice.cgi?action=factory.getCollect'
        )
        return ret.content.decode('utf-8')

    def _get_storage_values(self, *params):
        info = self.storage_device_info
        ret = []
        for param in params:
            try:
                ret.append(
                    re.search('.{}=([0-9.]+)'.format(param), info).group(1))
            except AttributeError:
                ret.append(None)
        if len(params) == 1:
            return ret[0]
        return ret

    @property
    def storage_used(self):
        return _express_as(self._get_storage_values(_USED), 'GB')

    @property
    def storage_total(self):
        return _express_as(self._get_storage_values(_TOTAL), 'GB')

    @property
    def storage_used_percent(self):
        used, total = self._get_storage_values(_USED, _TOTAL)
        try:
            return percent(used, total)
        except (TypeError, ValueError, ZeroDivisionError):
            return 'unknown'

    @property
    def storage_all(self):
        used, total = self._get_storage_values(_USED, _TOTAL)
        try:
            used_percent = percent(used, total)
        except (TypeError, ValueError, ZeroDivisionError):
            used_percent = 'unknown'
        return {
            'used_percent': used_percent,
            'used': _express_as(used, 'GB'),
            'total': _express_as(total, 'GB')}



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import logging
import re
import threading

import requests

from .exceptions import CommError, LoginError
from .utils import clean_url, pretty

from .audio import Audio
from .event import Event
from .log import Log
from .motion_detection import MotionDetection
from .nas import Nas
from .network import Network
from .ptz import Ptz
from .record import Record
from .snapshot import Snapshot
from .special import Special
from .storage import Storage
from .system import System
from .user_management import UserManagement
from .video import Video

from .config import TIMEOUT_HTTP_PROTOCOL, MAX_RETRY_HTTP_CONNECTION

_LOGGER = logging.getLogger(__name__)


# pylint: disable=too-many-ancestors
class Http(System, Network, MotionDetection, Snapshot,
           UserManagement, Event, Audio, Record, Video,
           Log, Ptz, Special, Storage, Nas):

    def __init__(self, host, port, user,
                 password, verbose=True, protocol='http', ssl_verify=True,
                 retries_connection=None, timeout_protocol=None):

        self._token_lock = threading.Lock()
        self._host = clean_url(host)
        self._port = port
        self._user = user
        self._password = password
        self._verbose = verbose
        self._protocol = protocol
        self._verify = ssl_verify
        self._base_url = self.__base_url()

        self._retries_default = (
            retries_connection if retries_connection is not None
            else MAX_RETRY_HTTP_CONNECTION)
        self._timeout_default = timeout_protocol or TIMEOUT_HTTP_PROTOCOL

        self._token = None
        self._name = None
        self._serial = None
        # Ignore comm errors in case camera happens to be off or there are
        # temporary network issues. User can retry later and we'll get token
        # then if camera is accessible again.
        try:
            self._generate_token()
        except CommError:
            pass

    def _generate_token(self):
        """Create authentation to use with requests."""
        cmd = 'magicBox.cgi?action=getMachineName'
        _LOGGER.debug('%s Trying Basic Authentication', self)
        self._token = requests.auth.HTTPBasicAuth(self._user, self._password)
        try:
            resp = self._command(cmd).content.decode('utf-8')
        except LoginError:
            _LOGGER.debug('%s Trying Digest Authentication', self)
            self._token = requests.auth.HTTPDigestAuth(
                self._user, self._password)
            try:
                resp = self._command(cmd).content.decode('utf-8')
            except LoginError as error:
                self._token = None
                raise error
        except CommError:
            self._token = None
            raise

        # check if user passed
        result = resp.lower()
        if 'invalid' in result or 'error' in result:
            _LOGGER.debug('%s Result from camera: %s', self,
                          resp.strip().replace('\r\n', ': '))
            self._token = None
            raise LoginError('Invalid credentials')

        self._name = pretty(resp.strip())

        _LOGGER.debug('%s Retrieving serial number', self)
        self._serial = pretty(self._command(
            'magicBox.cgi?action=getSerialNo').content.decode('utf-8').strip())

    def __repr__(self):
        """Default object representation."""
        return "<{0}:{1}>".format(self._name, self._serial)

    def as_dict(self):
        """Callback for __dict__."""
        cdict = self.__dict__.copy()
        redacted = '**********'
        cdict['_token'] = redacted
        cdict['_password'] = redacted
        return cdict

    # Base methods
    def __base_url(self, param=""):
        return '%s://%s:%s/cgi-bin/%s' % (self._protocol, self._host,
                                          str(self._port), param)

    def get_base_url(self):
        return self._base_url

    def command(self, cmd, retries=None, timeout_cmd=None, stream=False):
        """
        Args:
            cmd - command to execute via http
            retries - maximum number of retries each connection should attempt
            timeout_cmd - timeout
            stream - if True do not download entire response immediately
        """
        with self._token_lock:
            if not self._token:
                self._generate_token()
        return self._command(cmd, retries, timeout_cmd, stream)

    def _command(self, cmd, retries=None, timeout_cmd=None, stream=False):
        session = requests.Session()
        session.verify = self._verify
        url = self.__base_url(cmd)
        if retries is None:
            retries = self._retries_default
        for loop in range(1, 2 + retries):
            _LOGGER.debug("%s Running query attempt %s", self, loop)
            try:
                resp = session.get(
                    url,
                    auth=self._token,
                    stream=stream,
                    timeout=timeout_cmd or self._timeout_default,
                )
                if resp.status_code == 401:
                    raise LoginError
                resp.raise_for_status()
            except requests.RequestException as error:
                msg = re.sub(r'at 0x[0-9a-fA-F]+', 'at ADDRESS', repr(error))
                if loop > retries:
                    _LOGGER.debug(
                        "%s Query failed due to error: %s", self, msg)
                    raise CommError(error)
                _LOGGER.warning("%s Trying again due to error: %s", self, msg)
                continue
            else:
                break

        _LOGGER.debug(
            "%s Query worked. Exit code: <%s>", self, resp.status_code)
        return resp

    def command_audio(self, cmd, file_content, http_header,
                      timeout=None):
        with self._token_lock:
            if not self._token:
                self._generate_token()
        url = self.__base_url(cmd)
        try:
            requests.post(
                url,
                files=file_content,
                auth=self._token,
                headers=http_header,
                timeout=timeout or self._timeout_default
            )
        except requests.exceptions.ReadTimeout:
            pass



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import shutil
from urllib3.exceptions import HTTPError

from .exceptions import CommError


class Snapshot(object):
    def __get_config(self, config_name):
        ret = self.command(
            'configManager.cgi?action=getConfig&name={0}'.format(config_name)
        )
        return ret.content.decode('utf-8')

    @property
    def snapshot_config(self):
        return self.__get_config('Snap')

    def snapshot(self, channel=None, path_file=None, timeout=None):
        """
        Args:

            channel:
                Video input channel number

                If no channel param is used, don't send channel parameter
                so camera will use its default channel

            path_file:
                If path_file is provided, save the snapshot
                in the path

        Return:
            raw from http request
        """
        cmd = "snapshot.cgi"
        if channel is not None:
            cmd += "?channel={}".format(channel)
        ret = self.command(cmd, timeout_cmd=timeout, stream=True)

        if path_file:
            try:
                with open(path_file, 'wb') as out_file:
                    shutil.copyfileobj(ret.raw, out_file)
            except HTTPError as error:
                _LOGGER.debug(
                    "%s Snapshot to file failed due to error: %s",
                    self, repr(error))
                raise CommError(error)

        return ret.raw



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
from amcrest.utils import str2bool


class MotionDetection(object):
    def __get_config(self, config_name):
        ret = self.command(
            'configManager.cgi?action=getConfig&name={0}'.format(config_name)
        )
        return ret.content.decode('utf-8')

    @property
    def motion_detection(self):
        return self.__get_config("MotionDetect")

    def is_motion_detector_on(self):
        ret = self.motion_detection
        status = [s for s in ret.split() if '.Enable=' in s][0]\
            .split('=')[-1]
        return str2bool(status)  # pylint: disable=no-value-for-parameter

    def is_record_on_motion_detection(self):
        ret = self.motion_detection
        status = [s for s in ret.split() if '.RecordEnable=' in s][0]\
            .split('=')[-1]
        return str2bool(status)  # pylint: disable=no-value-for-parameter

    @motion_detection.setter
    def motion_detection(self, opt):
        if opt.lower() == "true" or opt.lower() == "false":
            ret = self.command(
                'configManager.cgi?action='
                'setConfig&MotionDetect[0].Enable={0}'.format(opt.lower())
            )
            if "ok" in ret.content.decode('utf-8').lower():
                return True

        return False

    @motion_detection.setter
    def motion_recording(self, opt):
        if opt.lower() == "true" or opt.lower() == "false":
            ret = self.command(
                'configManager.cgi?action='
                'setConfig&MotionDetect[0].EventHandler.RecordEnable={0}'
                .format(opt.lower())
            )
            if "ok" in ret.content.decode('utf-8').lower():
                return True

        return False



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
import shutil
from urllib3.exceptions import HTTPError

from .exceptions import CommError


class Special(object):

    def realtime_stream(self, channel=1, typeno=0, path_file=None):
        """
        If the stream is redirect to a file, use mplayer tool to
        visualize the video record

        camera.realtime_stream(path_file="/home/user/Desktop/myvideo)
        $ mplayer /home/user/Desktop/myvideo
        """
        ret = self.command(
            'realmonitor.cgi?action=getStream&channel={0}&subtype={1}'.format(
                channel, typeno), stream=True
        )

        if path_file:
            try:
                with open(path_file, 'wb') as out_file:
                    shutil.copyfileobj(ret.raw, out_file)
            except HTTPError as error:
                _LOGGER.debug("%s Realtime stream capture to file failed due "
                              "to error: %s", self, repr(error))
                raise CommError(error)

        return ret.raw

    def rtsp_url(self, channelno=None, typeno=None):
        """
        Return RTSP streaming url

        Params:
            channelno: integer, the video channel index which starts from 1,
                       default 1 if not specified.

            typeno: the stream type, default 0 if not specified. It can be
                    the following value:

                    0-Main Stream
                    1-Extra Stream 1 (Sub Stream)
                    2-Extra Stream 2 (Sub Stream)
        """
        if channelno is None:
            channelno = 1

        if typeno is None:
            typeno = 0

        cmd = 'cam/realmonitor?channel={0}&subtype={1}'.format(
            channelno, typeno)

        try:
            port = ':' + [x.split('=')[1] for x in self.rtsp_config.split()
                          if x.startswith('table.RTSP.Port=')][0]
        except IndexError:
            port = ''
        return 'rtsp://{}:{}@{}{}/{}'.format(
            self._user, self._password, self._host, port, cmd)

    # pylint: disable=pointless-string-statement
    """
    11/05/2016

    Looks like the mjpeg support was removed from the mainstream in one of
    the firmware updates, so subtype=0 is not working.

    I have tested with the below firmware and still not working
    in the mainstream:
        Amcrest_IPC-AWXX_Eng_N_V2.420.AC00.15.R.20160908.bin

    Based on above, the cmd we send to camera should use: subtype=1 until
    mainstream is back.

    Simple translation:
        subtype=0 for mainstream
        subtype=1 for substream

    Tested on device: IP2M-841B

    Users complaining here too:
    https://amcrest.com/forum/technical-discussion-f3/lost-mjpeg-encode-for-main-stream-after-firmware-u-t1516.html
    """

    def mjpeg_url(self, channelno=None, typeno=None):
        """
        Return MJPEG streaming url

        Params:
            channelno: integer, the video channel index which starts from 1,
                       default 1 if not specified.

            typeno: the stream type, default 0 if not specified. It can be
                    the following value:

                    0-Main Stream
                    1-Extra Stream 1 (Sub Stream)
                    2-Extra Stream 2 (Sub Stream)
        """
        if channelno is None:
            channelno = 0

        if typeno is None:
            typeno = 1

        cmd = "mjpg/video.cgi?channel={0}&subtype={1}".format(
            channelno, typeno)
        return '{0}{1}'.format(self._base_url, cmd)

    def mjpg_stream(self, channelno=None, typeno=None, path_file=None):
        """
        Params:
            channelno: integer, the video channel index which starts from 1,
                       default 1 if not specified.

            typeno: the stream type, default 0 if not specified. It can be
                    the following value:

                    0-Main Stream
                    1-Extra Stream 1 (Sub Stream)
                    2-Extra Stream 2 (Sub Stream)
        """
        cmd = self.mjpeg_url(channelno=channelno, typeno=typeno)
        ret = self.command(cmd, stream=True)

        if path_file:
            try:
                with open(path_file, 'wb') as out_file:
                    shutil.copyfileobj(ret.raw, out_file)
            except HTTPError as error:
                _LOGGER.debug(
                    "%s MJPEG stream capture to file failed due to error: %s",
                    self, repr(error))
                raise CommError(error)

        return ret.raw



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# vim:sw=4:ts=4:et
from . import utils


class Video(object):

    @property
    def video_max_extra_stream(self):
        ret = self.command(
            'magicBox.cgi?action=getProductDefinition&name=MaxExtraStream'
        )
        return ret.content.decode('utf-8')

    @property
    def video_color_config(self):
        ret = self.command(
            'configManager.cgi?action=getConfig&name=VideoColor'
        )
        return ret.content.decode('utf-8')

    @property
    def encode_capability(self):
        ret = self.command(
            'encode.cgi?action=getCaps'
        )
        return ret.content.decode('utf-8')

    def encode_config_capability(self, channel):
        ret = self.command(
            'encode.cgi?action=getConfigCaps&channel={0}'.format(channel)
        )
        return ret.content.decode('utf-8')

    @property
    def encode_media(self):
        ret = self.command(
            'configManager.cgi?action=getConfig&name=Encode'
        )
        return ret.content.decode('utf-8')

    @property
    def encode_region_interested(self):
        ret = self.command(
            'configManager.cgi?action=getConfig&name=VideoEncodeROI'
        )
        return ret.content.decode('utf-8')

    @property
    def video_channel_title(self):
        ret = self.command(
            'configManager.cgi?action=getConfig&name=ChannelTitle'
        )
        return ret.content.decode('utf-8')

    # pylint: disable=invalid-name
    @property
    def video_input_channels_device_supported(self):
        ret = self.command(
            'devVideoInput.cgi?action=getCollect'
        )
        return ret.content.decode('utf-8')

    # pylint: disable=invalid-name
    @property
    def video_output_channels_device_supported(self):
        ret = self.command(
            'devVideoOutput.cgi?action=getCollect'
        )
        return ret.content.decode('utf-8')

    # pylint: disable=invalid-name
    @property
    def video_max_remote_input_channels(self):
        ret = self.command(
            'magicBox.cgi?action=getProductDefinition&name='
            'MaxRemoteInputChannels'
        )
        return ret.content.decode('utf-8')

    @property
    def video_standard(self):
        ret = self.command(
            'configManager.cgi?action=getConfig&name=VideoStandard'
        )
        return ret.content.decode('utf-8')

    @video_standard.setter
    def video_standard(self, std):
        ret = self.command(
            'configManager.cgi?action=setConfig&VideoStandard={0}'.format(std)
        )
        return ret.content.decode('utf-8')

    @property
    def video_widget_config(self):
        ret = self.command(
            'configManager.cgi?action=getConfig&name=VideoWidget'
        )
        return ret.content.decode('utf-8')

    def video_input_capability(self, channel):
        ret = self.command(
            'devVideoInput.cgi?action=getCaps&channel={0}'.format(channel)
        )
        return ret.content.decode('utf-8')

    def video_coordinates_current_window(self, channel):
        ret = self.command(
            'devVideoInput.cgi?action='
            'getCurrentWindow&channel={0}'.format(channel)
        )
        return ret.content.decode('utf-8')

    @property
    def video_in_options(self):
        ret = self.command(
            'configManager.cgi?action=getConfig&name=VideoInOptions'
        )
        return ret.content.decode('utf-8')

    def video_in_option(self, param, profile='Day'):
        """
        Return video input option.

        Params:
            param - parameter, such as 'DayNightColor'
            profile - 'Day', 'Night' or 'Normal'
        """
        if profile == 'Day':
            field = param
        else:
            field = '{}Options.{}'.format(profile, param)
        return utils.pretty(
            [opt for opt in self.video_in_options.split()
             if '].{}='.format(field) in opt][0])

    def set_video_in_option(self, param, value, profile='Day'):
        if profile == 'Day':
            field = param
        else:
            field = '{}Options.{}'.format(profile, param)
        ret = self.command(
            'configManager.cgi?action=setConfig'
            '&VideoInOptions[0].{}={}'.format(field, value)
        )
        return ret.content.decode('utf-8')

    @property
    def day_night_color(self):
        """
        Return Day & Night Color Mode for Day profile.

        Result is 0: always multicolor
                  1: autoswitch along with brightness
                  2: always monochrome
        """
        return int(self.video_in_option('DayNightColor'))

    @day_night_color.setter
    def day_night_color(self, value):
        return self.set_video_in_option('DayNightColor', value)

    @property
    def smart_ir(self):
        """Return if SmartIR is on."""
        return self.video_in_option('InfraRed') == 'false'

    @smart_ir.setter
    def smart_ir(self, value):
        # It's not clear why from the HTTP API SDK doc, but setting InfraRed
        # to false sets the Night Vision Mode to SmartIR, whereas setting it
        # to true sets the Night Vision Mode to OFF. Night Vision Mode has a
        # third setting of Manual, but that must be selected some other way
        # via the HTTP API.
        return self.set_video_in_option('InfraRed', str(not value).lower())

    @property
    def video_out_options(self):
        ret = self.command(
            'configManager.cgi?action=getConfig&name=VideoOut'
        )
        return ret.content.decode('utf-8')

    @property
    def video_enabled(self):
        """Return if any video stream enabled."""
        return utils.extract_audio_video_enabled('Video', self.encode_media)

    @video_enabled.setter
    def video_enabled(self, enable):
        self.command(utils.enable_audio_video_cmd('Video', enable))
        self.smart_ir = enable
